Imports System.Collections.Generic
Imports System.Data.SqlClient
Imports System.text
Imports System.io

Public Class MainForm

#Region "Constants and module-level vars"

    Private Const INVOKE_RANDOM_COUNT As Integer = 1000

    Private mdbconn As SqlConnection
    Private mdtGrid As DataTable
    Private mcmdGetRate As SqlCommand
    Private moRateCache As New Dictionary(Of String, MoneyWithError)
    Private mex As Exception
    Private miRetrieveStart As Integer
    Private miWait As Integer
    Private miCumAvgSum As Integer = 0
    Private miCumAvgCnt As Integer = 0

#End Region

#Region "Internal methods"

    ''' <summary>
    ''' Populates mdtGrid with either an empty DataTable, or filled based on CustomerService records in database.
    ''' </summary>
    ''' <param name="bEmpty">True causes empty result set.</param>
    ''' <remarks></remarks>
    Private Sub RetrieveServices(ByVal bEmpty As Boolean)
        Dim da As New SqlDataAdapter(GetRetrieveCommand(bEmpty))
        mdtGrid = New DataTable
        da.Fill(mdtGrid)
        da.Dispose()
    End Sub

    Private Function GetRetrieveCommand(ByVal bEmpty As Boolean) As SqlCommand
        Dim cmd As New SqlCommand("up_RetrieveServices", mdbconn)
        cmd.CommandType = CommandType.StoredProcedure
        cmd.CommandTimeout = mdbconn.ConnectionTimeout
        cmd.Parameters.AddWithValue("EmptySet", bEmpty)
        Return cmd
    End Function

    ''' <summary>
    ''' Returns a rate amount, either from memory cache or database.  Returns as a class that can indicate an exact rate amount or an error condition.
    ''' </summary>
    ''' <param name="oCustomerID"></param>
    ''' <param name="oServiceTypeID"></param>
    ''' <param name="oDelSubLocationID"></param>
    ''' <param name="oContainerSizeID"></param>
    ''' <param name="oWeightClassID"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function GetRateAmount(ByVal oCustomerID As Nullable(Of Integer), ByVal oServiceTypeID As Nullable(Of Integer), ByVal oDelSubLocationID As Nullable(Of Integer), ByVal oContainerSizeID As Nullable(Of Integer), ByVal oWeightClassID As Nullable(Of Integer), ByVal oServiceDate As Nullable(Of DateTime)) As MoneyWithError
        If Not oCustomerID.HasValue Then
            Return New MoneyWithError("Must specify a customer.")
        End If

        If Not oServiceTypeID.HasValue Then
            Return New MoneyWithError("Must specify a service type.")
        End If

        Dim oRate As MoneyWithError = Nothing
        Dim sKey As String = GetRateCacheKey(oCustomerID.Value, oServiceTypeID.Value, oDelSubLocationID, oContainerSizeID, oWeightClassID)

        If Not moRateCache.TryGetValue(sKey, oRate) Then
            ' Not in local cache, must retrieve from database
            oRate = GetRateAmountFromDatabase(oCustomerID.Value, oServiceTypeID.Value, oDelSubLocationID, oContainerSizeID, oWeightClassID, oServiceDate)

            ' Add to cache for future attempts
            moRateCache.Add(sKey, oRate)
        End If

        Return oRate
    End Function

    ''' <summary>
    ''' Invokes the stored procedure we're using to retrieve the best matching rate amount, or an error if none or multiples found.
    ''' </summary>
    ''' <param name="iCustomerID"></param>
    ''' <param name="iServiceTypeID"></param>
    ''' <param name="oDelSubLocationID"></param>
    ''' <param name="oContainerSizeID"></param>
    ''' <param name="oWeightClassID"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function GetRateAmountFromDatabase(ByVal iCustomerID As Integer, ByVal iServiceTypeID As Integer, ByVal oDelSubLocationID As Nullable(Of Integer), ByVal oContainerSizeID As Nullable(Of Integer), ByVal oWeightClassID As Nullable(Of Integer), ByVal oServiceDate As Nullable(Of DateTime)) As MoneyWithError
        If mcmdGetRate Is Nothing Then
            mcmdGetRate = New SqlCommand("up_GetRateByCriteria", mdbconn)
            mcmdGetRate.CommandType = CommandType.StoredProcedure
            With mcmdGetRate.Parameters
                .Add("CustomerID", SqlDbType.Int)
                .Add("ServiceTypeID", SqlDbType.Int)
                .Add("ServiceDate", SqlDbType.DateTime)
                .Add("DeliverySubLocationID", SqlDbType.Int)
                .Add("ContainerSizeID", SqlDbType.Int)
                .Add("WeightClassID", SqlDbType.Int)
            End With
        End If

        With mcmdGetRate.Parameters
            .Item("CustomerID").Value = iCustomerID
            .Item("ServiceTypeID").Value = iServiceTypeID
            .Item("ServiceDate").Value = oServiceDate.GetValueOrDefault(DateTime.Now)
            .Item("DeliverySubLocationID").Value = oDelSubLocationID.GetValueOrDefault(-1)
            .Item("ContainerSizeID").Value = oContainerSizeID.GetValueOrDefault(-1)
            .Item("WeightClassID").Value = oWeightClassID.GetValueOrDefault(-1)
        End With

        Dim drd As SqlDataReader = mcmdGetRate.ExecuteReader(CommandBehavior.SingleResult)
        Dim iCnt As Integer = 0
        Dim dValue As Decimal = 0

        Try
            While drd.Read()
                dValue = drd.GetDecimal(0)
                iCnt += 1
            End While

        Catch ex As Exception
            Throw

        Finally
            drd.Close()
        End Try

        If iCnt > 1 Then
            Return New MoneyWithError("Multiple matching rates found (" & iCnt & " matches).")
        Else
            If iCnt = 0 Then
                Return New MoneyWithError("No rate found.")
            Else
                Return New MoneyWithError(dValue)
            End If
        End If
    End Function

    ''' <summary>
    ''' Returns a key into our local memory cache, given all possible inputs.
    ''' </summary>
    ''' <param name="iCustomerID"></param>
    ''' <param name="iServiceTypeID"></param>
    ''' <param name="oDelSubLocationID"></param>
    ''' <param name="oContainerSizeID"></param>
    ''' <param name="oWeightClassID"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function GetRateCacheKey(ByVal iCustomerID As Integer, ByVal iServiceTypeID As Integer, ByVal oDelSubLocationID As Nullable(Of Integer), ByVal oContainerSizeID As Nullable(Of Integer), ByVal oWeightClassID As Nullable(Of Integer)) As String
        Dim sb As New StringBuilder
        sb.Append(iCustomerID)
        sb.Append("-")
        sb.Append(iServiceTypeID)
        sb.Append("-")

        If oDelSubLocationID.HasValue Then
            sb.Append(oDelSubLocationID)
        Else
            sb.Append("~")
        End If
        sb.Append("-")

        If oContainerSizeID.HasValue Then
            sb.Append(oContainerSizeID)
        Else
            sb.Append("~")
        End If
        sb.Append("-")

        If oWeightClassID.HasValue Then
            sb.Append(oWeightClassID)
        Else
            sb.Append("~")
        End If

        Return sb.ToString
    End Function

    Private Sub HandleException(ByVal ex As Exception)
        MessageBox.Show("An unexpected error occurred:" & vbCrLf & ex.Message, "Error encountered.", MessageBoxButtons.OK, MessageBoxIcon.Error)
    End Sub

    Private Function GetNullableInt(ByVal val As Object) As Nullable(Of Integer)
        If val Is Nothing OrElse val Is DBNull.Value Then
            Return New Nullable(Of Integer)
        Else
            Dim out As New Nullable(Of Integer)
            out = CInt(val)
            Return out
        End If
    End Function

    Private Function GetNullableDate(ByVal val As Object) As Nullable(Of DateTime)
        If val Is Nothing OrElse val Is DBNull.Value Then
            Return New Nullable(Of DateTime)
        Else
            Dim out As New Nullable(Of DateTime)
            out = CDate(val)
            Return out
        End If
    End Function

    ''' <summary>
    ''' Invoked on a worker thread, performs long-running data retrieval.
    ''' </summary>
    ''' <param name="state"></param>
    ''' <remarks></remarks>
    Private Sub InternalRetrieveServices(ByVal state As Object)
        Try
            RetrieveServices(False)
            Me.Invoke(New MethodInvoker(AddressOf FinishRetrieveServices))

        Catch ex As Exception
            mex = ex
            Me.Invoke(New MethodInvoker(AddressOf FinishRetrieveServicesError))
        End Try
    End Sub

    ''' <summary>
    ''' On successful completion of data retrieval, renders the results.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub FinishRetrieveServices()
        Try
            tsbRetrieve.Enabled = True
            tsbRandom.Enabled = True
            timRetrieve.Enabled = False

            bsGrid.DataSource = mdtGrid
            bsGrid.ResetBindings(False)

            tsslStatus.Text = "Retrieved " & mdtGrid.Rows.Count.ToString("#,##0") & " rows in " & (Environment.TickCount - miRetrieveStart).ToString("#,##0") & " milliseconds, including rendering."
            Me.Cursor = Cursors.Default

        Catch ex As Exception
            HandleException(ex)
        End Try
    End Sub

    ''' <summary>
    ''' On a failure during data retrieval, renders the error.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub FinishRetrieveServicesError()
        tsbRetrieve.Enabled = True
        tsbRandom.Enabled = True
        timRetrieve.Enabled = False
        Me.Cursor = Cursors.Default
        tsslStatus.Text = "Error encountered."
        HandleException(mex)
    End Sub

#End Region

#Region "Event handlers"

    Private Sub tsbRetrieve_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tsbRetrieve.Click
        Try
            tsbRetrieve.Enabled = False
            tsbRandom.Enabled = False
            timRetrieve.Enabled = True
            miWait = 0
            Me.Cursor = Cursors.AppStarting
            miRetrieveStart = Environment.TickCount
            Threading.ThreadPool.QueueUserWorkItem(AddressOf InternalRetrieveServices)

        Catch ex As Exception
            HandleException(ex)
        End Try
    End Sub

    Private Sub MainForm_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles Me.FormClosed
        Try
            If Not mdbconn Is Nothing Then
                mdbconn.Dispose()
            End If

        Catch ex As Exception
            HandleException(ex)
        End Try
    End Sub

    Private Sub MainForm_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        e.Cancel = timRetrieve.Enabled
    End Sub

    Private Sub MainForm_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        Try
            dgvServices.AutoGenerateColumns = False

            ' Attempt to open connection for duration of app
            Dim sDBFile As String = Application.StartupPath & "\..\data\RateSample.mdf"
            mdbconn = New SqlConnection("Server=.\SQLExpress;Connect Timeout=60;AttachDbFilename=" & sDBFile & ";Initial Catalog=RateSample;Trusted_Connection=Yes")
            mdbconn.Open()

            Dim dbconn As New SqlConnection(mdbconn.ConnectionString)
            Try
                ' We do this for performance reasons:  otherwise, the first retrieval can be very, very slow since we're using tempdb
                ' The "default" tempdb for SQLExpress is quite small and with 10% growth, kills performance
                ' Note that we do this on its own connection - if it bombs, its possible the connection can become "confused"
                dbconn.Open()
                Dim cmd As New SqlCommand("ALTER DATABASE tempdb MODIFY FILE (NAME = tempdev, SIZE = 50); ALTER DATABASE tempdb MODIFY FILE (NAME = templog, SIZE = 50);", dbconn)
                cmd.ExecuteNonQuery()

            Catch ex As Exception
                ' Can get legit errors to be ignored

            Finally
                dbconn.Dispose()
            End Try

            Me.ServiceTypeTableAdapter.Connection = mdbconn
            Me.ServiceTypeTableAdapter.Fill(Me.LookupsData.ServiceType)
            Me.CustomerTableAdapter.Connection = mdbconn
            Me.CustomerTableAdapter.Fill(Me.LookupsData.Customer)
            Me.SubLocationTableAdapter.Connection = mdbconn
            Me.SubLocationTableAdapter.Fill(Me.LookupsData.SubLocation)
            Me.ContainerSizeTableAdapter.Connection = mdbconn
            Me.ContainerSizeTableAdapter.Fill(Me.LookupsData.ContainerSize)
            Me.WeightClassTableAdapter.Connection = mdbconn
            Me.WeightClassTableAdapter.Fill(Me.LookupsData.WeightClass)

            ' On screen open, we have an empty, editable grid
            RetrieveServices(True)
            bsGrid.DataSource = mdtGrid
            bsGrid.ResetBindings(False)

        Catch ex As Exception
            HandleException(ex)
            Me.Close()
        End Try
    End Sub

    Private Sub dgvServices_CellValueChanged(ByVal sender As System.Object, ByVal e As System.Windows.Forms.DataGridViewCellEventArgs) Handles dgvServices.CellValueChanged
        Try
            If e.RowIndex >= 0 AndAlso e.ColumnIndex <> dgvServices.Columns("RateAmount").Index Then
                tsslStatus.Text = "Retrieving Rate..."
                Dim iStart As Integer = Environment.TickCount

                Dim dgr As DataGridViewRow = dgvServices.Rows(e.RowIndex)
                Dim oRate As MoneyWithError = GetRateAmount( _
                    GetNullableInt(dgr.Cells("CustomerID").Value) _
                    , GetNullableInt(dgr.Cells("ServiceTypeID").Value) _
                    , GetNullableInt(dgr.Cells("DeliverySubLocationID").Value) _
                    , GetNullableInt(dgr.Cells("ContainerSizeID").Value) _
                    , GetNullableInt(dgr.Cells("WeightClassID").Value) _
                    , GetNullableDate(dgr.Cells("ServiceDate").Value))

                ' Update the rate amount, set/clear error if present
                If oRate.IsError Then
                    dgr.ErrorText = oRate.Message
                    dgr.Cells("RateAmount").Value = DBNull.Value
                Else
                    dgr.ErrorText = String.Empty
                    dgr.Cells("RateAmount").Value = oRate.MoneyValue
                End If

                If dgr.Cells("ServiceDate").Value.ToString = String.Empty Then
                    dgr.Cells("ServiceDate").Value = DateTime.Today
                End If

                Dim iTime As Integer = (Environment.TickCount - iStart)

                If e.ColumnIndex > 0 OrElse dgr.ErrorText = String.Empty Then
                    miCumAvgSum += iTime
                    miCumAvgCnt += 1
                    tsslStatus.Text = "Retrieved Rate in " & iTime.ToString("#,##0") & " milliseconds, cumulative average " & (CDbl(miCumAvgSum) / miCumAvgCnt).ToString("#,##0") & " milliseconds"
                Else
                    tsslStatus.Text = String.Empty
                End If
            End If

        Catch ex As Exception
            tsslStatus.Text = "Error encountered."
            HandleException(ex)
        End Try
    End Sub

    Private Sub timRetrieve_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles timRetrieve.Tick
        Dim sText As String = "Retrieving data"
        tsslStatus.Text = sText.PadRight(sText.Length + miWait, "."c)
        miWait = (miWait + 1) Mod 20
    End Sub

    Private Sub tsbRandom_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tsbRandom.Click
        Try
            Me.Cursor = Cursors.WaitCursor
            Dim rnd As New System.Random(87581232)
            Dim iStart As Integer = Environment.TickCount

            For iCnt As Integer = 1 To INVOKE_RANDOM_COUNT
                GetRateAmountFromDatabase(rnd.Next(1, Me.LookupsData.Customer.Rows.Count) _
                    , rnd.Next(1, Me.LookupsData.ServiceType.Rows.Count) _
                    , rnd.Next(1, Me.LookupsData.SubLocation.Rows.Count) _
                    , rnd.Next(1, Me.LookupsData.ContainerSize.Rows.Count) _
                    , rnd.Next(1, Me.LookupsData.WeightClass.Rows.Count) _
                    , Nothing)
            Next

            tsslStatus.Text = "Invoked get rate " & INVOKE_RANDOM_COUNT.ToString("#,##0") & " times in " & (Environment.TickCount - iStart).ToString("#,##0") & " milliseconds"
            Me.Cursor = Cursors.Default

        Catch ex As Exception
            Me.Cursor = Cursors.Default
            HandleException(ex)
        End Try
    End Sub

#End Region

End Class
